using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;

using Microsoft.Kinect;
using System.IO;
using KinectUtils;


// Note: This version of the game uses the Kinect Manager class to interact with the Kinect sensor
//        You can find our more about this class in Chapter 10. 

namespace KinectBlockBuster
{
    /// <summary>
    /// Holds information about a Block in the wall
    /// </summary>
    struct BlockSprite
    {
        /// <summary>
        /// The texture used to draw the sprite
        /// </summary>
        public Texture2D SpriteTexture;

        /// <summary>
        /// The rectangle on the screen where the sprite is placed
        /// </summary>
        public Rectangle SpriteRectangle;

        /// <summary>
        /// The score gained when the Block is hit
        /// </summary>
        public int SpriteScore;

        /// <summary>
        /// The colour of the Block
        /// </summary>
        public Color BlockColour;

        /// <summary>
        /// True if the sprite is visible on the screen
        /// </summary>
        public bool IsVisible;

    }


    class KinectController
    {
        KinectManager myKinectMgr;
        KinectSensor myKinect;

        public bool SetupKinect()
        {
            try
            {
                myKinectMgr = new KinectManager(
                    0, // Sensor number 0
                    ColorImageFormat.Undefined,                 // No color image needed
                    DepthImageFormat.Resolution640x480Fps30,    // Depth image needed
                    false);                                      // Skeleton image needed

                myKinect = myKinectMgr.Kinect;

                myKinectMgr.DoProcessDepthFrame +=
                    new KinectManager.ProcessDepthFrame(processDepthFrame);

                myKinectMgr.Start();
                return true;
            }
            catch
            {
                ErrorMessage = "Kinect initialise failed";
                return false;
            }
        }

        public void ShutdownKinect()
        {
            if (myKinectMgr != null)
                myKinectMgr.Stop();
        }

        private GraphicsDevice graphicsDevice;

        public string ErrorMessage;

        public KinectController(GraphicsDevice graphicsDevice)
        {
            this.graphicsDevice = graphicsDevice;
            ErrorMessage = "";
        }

        const int slotSizeFactor = 7;
        const int noOfSlots = 4096 >> slotSizeFactor;

        int[] SlotCounts = new int[noOfSlots];

        byte[] depthColorImage = null;

        int xPosition;
        int xTotal;
        float halfWidth;
        int width;

        /// <summary>
        /// Calculates and returns the X controller value
        /// -1 means hard left and +1 means hard right
        /// 0 means the controller is central
        /// </summary>
        public float X
        {
            get
            {
                return (xPosition - halfWidth) / halfWidth;
            }
        }

        void setColorImage(short[] depthData, int width, int height)
        {

            // Clear the depth counters in each slot
            for (int slotNo = 0; slotNo < SlotCounts.Length; slotNo++)
            {
                SlotCounts[slotNo] = 0;
            }

            // Make a pass through the image counting depth pixels

            for (int depthPos = 0; depthPos < depthData.Length; depthPos++)
            {
                int depthValue = depthData[depthPos] >> 3;
                // Check for the invalid values

                if (depthValue == myKinect.DepthStream.UnknownDepth ||
                    depthValue == myKinect.DepthStream.TooFarDepth ||
                    depthValue == myKinect.DepthStream.TooNearDepth)
                {
                    continue;
                }

                // Got a valid depth value
                // Find the slot and increase the counter

                int slotNumber = depthValue >> slotSizeFactor;
                SlotCounts[slotNumber]++;
            }

            // Now find the depth with the most pixels
            int maxCount = 0;
            int maxPos = 0;

            for (int slotNo = 0; slotNo < SlotCounts.Length; slotNo++)
            {
                if (SlotCounts[slotNo] > maxCount)
                {
                    maxCount = SlotCounts[slotNo];
                    maxPos = slotNo;
                }
            }

            // Now color those pixels red in the image

            int depthColorImagePos = 0;
            xTotal = 0;

            for (int depthPos = 0; depthPos < depthData.Length; depthPos++)
            {
                int depthValue = depthData[depthPos] >> 3;
                // Check for the invalid values

                if (depthValue == myKinect.DepthStream.UnknownDepth ||
                    depthValue == myKinect.DepthStream.TooFarDepth ||
                    depthValue == myKinect.DepthStream.TooNearDepth)
                {
                    // Color invalid pixels black
                    // Red
                    depthColorImage[depthColorImagePos] = 0;
                    depthColorImagePos++;
                    // Green
                    depthColorImage[depthColorImagePos] = 0;
                    depthColorImagePos++;
                    // Blue
                    depthColorImage[depthColorImagePos] = 0;
                    depthColorImagePos++;
                }
                else
                {
                    byte depthByte = (byte)(255 - (depthValue >> 4));

                    if (depthValue >> slotSizeFactor == maxPos)
                    {
                        // Red
                        depthColorImage[depthColorImagePos] = depthByte;
                        depthColorImagePos++;
                        // Green
                        depthColorImage[depthColorImagePos] = 0;
                        depthColorImagePos++;
                        // Blue
                        depthColorImage[depthColorImagePos] = 0;
                        depthColorImagePos++;

                        // Add the x position of this pixel to the total
                        xTotal = xTotal + (depthPos % width);

                    }
                    else
                    {
                        // Red
                        depthColorImage[depthColorImagePos] = depthByte;
                        depthColorImagePos++;
                        // Green
                        depthColorImage[depthColorImagePos] = depthByte;
                        depthColorImagePos++;
                        // Blue
                        depthColorImage[depthColorImagePos] = depthByte;
                        depthColorImagePos++;
                    }
                }
                // transparency
                depthColorImagePos++;
            }

            xPosition = xTotal / SlotCounts[maxPos];
        }

        public Texture2D DepthTexture = null;
        short[] depthData = null;

        void processDepthFrame(DepthImageFrame depthFrame)
        {
            if (depthData == null)
                depthData = new short[depthFrame.PixelDataLength];

            if (depthColorImage == null)
            {
                depthColorImage = new byte[depthFrame.PixelDataLength * 4];
                width = depthFrame.Width;
                halfWidth = depthFrame.Width / 2;
            }

            depthFrame.CopyPixelDataTo(depthData);

            setColorImage(depthData, depthFrame.Width, depthFrame.Height);
        }

        public void Update()
        {
            if (depthColorImage != null)
            {
                if ( DepthTexture == null )
                    DepthTexture = new Texture2D(graphicsDevice, 
                        myKinect.DepthStream.FrameWidth,
                        myKinect.DepthStream.FrameHeight);
                DepthTexture.SetData(depthColorImage);
            }
        }
    }


    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        KinectController kinect;

        // Game Data
        SpriteFont messageFont;

        Rectangle depthRectangle;
        Color depthColor = new Color(255, 255, 255);

        Texture2D ballTexture;
        Rectangle ballRectangle;

        float ballXpos;
        float ballYpos;
        float ballXspeed;
        float ballYspeed;
        float ballStartSpeed;

        Texture2D batTexture;
        Rectangle batRectangle;
        float batSpeed;

        Texture2D blockTexture;
        BlockSprite[] wall = new BlockSprite[20];
        int blocksLeft = 20;

        BlockSprite[] wall2 = new BlockSprite[15];
        int wall2BlocksLeft = 15;

        int wallYLimit;
        int wallYStart;
        int wallYPosition;
        int wallYStep;

        float ballSpeedUp;

        SoundEffect beep;
        SoundEffect gameOver;
        SoundEffect lose;

        int score;
        int lives = 3;

        enum GameState
        {
            error,
            attractMode,
            gamePlaying,
        };

        GameState state = GameState.attractMode;

        string message = "Welcome to Kinect BlockBuster\r\n         Press Space to begin";

        Color ballColor = Color.Yellow;

        GamePadState oldGamePadState;
        GamePadState gamePad1;

        KeyboardState keyState;
        KeyboardState oldKeyState;

        private Rectangle getScaleRectangle
            (Texture2D texture, float widthFactor)
        {
            float aspectRatio =
                (float)texture.Width / texture.Height;
            float width =
                GraphicsDevice.Viewport.Width / widthFactor;
            float height = width / aspectRatio;

            Rectangle result = new Rectangle(
                0,              // x position
                0,              // y position
                (int)width,     // width
                (int)height);   // height

            return result;
        }

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            kinect = new KinectController(GraphicsDevice);
            
            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // Load the sounds
            beep = Content.Load<SoundEffect>(@"Sounds\beep");
            gameOver = Content.Load<SoundEffect>(@"Sounds\gameOver");
            lose = Content.Load<SoundEffect>(@"Sounds\lose");

            // load the font
            messageFont = Content.Load<SpriteFont>("MessageFont");

            // Ball setup

            ballTexture = Content.Load<Texture2D>(@"Images\ball");
            ballRectangle = getScaleRectangle(ballTexture, 30);

            // put the ball near the middle of the screen
            ballXpos = GraphicsDevice.Viewport.Width / 2;
            ballYpos = GraphicsDevice.Viewport.Height / 2;

            // set the initial ball speed
            ballStartSpeed = GraphicsDevice.Viewport.Width / 400f;
            ballXspeed = ballStartSpeed;
            ballYspeed = ballStartSpeed;

            ballSpeedUp = ballXspeed / 5;

            // Bat setup

            batTexture = Content.Load<Texture2D>(@"Images\bat");
            batRectangle = getScaleRectangle(batTexture, 8);
            batRectangle.X = (GraphicsDevice.Viewport.Width - batRectangle.Width) / 2;
            batRectangle.Y = GraphicsDevice.Viewport.Height - (batRectangle.Height * 3);

            // set the initial bat speed
            batSpeed = GraphicsDevice.Viewport.Width / 100f;

            // Block setup

            blockTexture = Content.Load<Texture2D>(@"Images\block");

            float blockWidth = GraphicsDevice.Viewport.Width / (float)wall.Length;
            float blockHeight = GraphicsDevice.Viewport.Height / 30.0f;

            // set up the limit for the blocks
            wallYLimit = (int)(GraphicsDevice.Viewport.Height * 0.75f);
            wallYStart = (int)(GraphicsDevice.Viewport.Height * 0.2f);
            wallYPosition = wallYStart;
            wallYStep = (int)(GraphicsDevice.Viewport.Height * 0.02f);

            for (int i = 0; i < wall.Length; i++)
            {
                wall[i].SpriteRectangle = new Rectangle(
                    (int)((i * blockWidth) + 0.5),  // X position
                    wallYPosition,                  // Y position
                    (int)(blockWidth + 0.5),        // width
                    (int)(blockHeight + 0.5));     // height

                wall[i].SpriteTexture = blockTexture;
                wall[i].SpriteScore = 20;
                wall[i].BlockColour = Color.Red;
                wall[i].IsVisible = true;
            }

            blockWidth = GraphicsDevice.Viewport.Width / (float)wall2.Length;
            wallYPosition += (int)(blockHeight * 4);

            for (int i = 0; i < wall2.Length; i++)
            {
                wall2[i].SpriteRectangle = new Rectangle(
                    (int)((i * blockWidth) + 0.5),  // X position
                    wallYPosition,                  // Y position
                    (int)(blockWidth + 0.5),        // width
                    (int)(blockHeight + 0.5));     // height

                wall2[i].SpriteTexture = blockTexture;
                wall2[i].SpriteScore = 20;
                wall2[i].BlockColour = Color.Blue;
            }

            depthRectangle = new Rectangle(0, 0, GraphicsDevice.Viewport.Width, GraphicsDevice.Viewport.Height);

            if (!kinect.SetupKinect())
            {
                message = kinect.ErrorMessage;
                state = GameState.error;
            }
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        private void startGame()
        {
            ballXpos = GraphicsDevice.Viewport.Width / 2;
            ballYpos = GraphicsDevice.Viewport.Height / 2;
            ballXspeed = ballStartSpeed;
            ballYspeed = -ballStartSpeed;
            score = 0;
            lives = 3;
            for (int i = 0; i < wall.Length; i++)
            {
                wall[i].SpriteRectangle.Y = wallYPosition;
                wall[i].IsVisible = true;
            }

            for (int i = 0; i < wall2.Length; i++)
            {
                wall2[i].SpriteRectangle.Y = wallYPosition + (wall2[i].SpriteRectangle.Height * 4);
                wall2[i].IsVisible = true;
            }
            state = GameState.gamePlaying;
        }

        private void loseLife()
        {
            lives--;

            if (lives <= 0)
            {
                gameOver.Play();
                message = "Game Score: " + score.ToString();
                state = GameState.attractMode;
            }
            else
            {
                lose.Play();
            }
        }

        private void ballMotionUpdate()
        {
            ballXpos += ballXspeed;
            ballRectangle.X = (int)(ballXpos + 0.5f);

            if (((ballRectangle.X + ballRectangle.Width) >
                 GraphicsDevice.Viewport.Width) ||
                 (ballRectangle.X < 0))
            {
                ballXspeed = ballXspeed * -1;
            }

            ballYpos += ballYspeed;
            ballRectangle.Y = (int)(ballYpos + 0.5f);

            if (((ballRectangle.Y + ballRectangle.Height) >
                 GraphicsDevice.Viewport.Height) ||
                 (ballRectangle.Y < 0))
            {
                ballYspeed = ballYspeed * -1;
            }
        }

        private void attractUpdate()
        {
            ballMotionUpdate();

            if ((gamePad1.Buttons.Start == ButtonState.Pressed &&
              oldGamePadState.Buttons.Start == ButtonState.Released) ||
                keyState.IsKeyDown(Keys.Space) && oldKeyState.IsKeyUp(Keys.Space))
            {
                startGame();
            }
        }

        private void gameUpdate()
        {
            ballMotionUpdate();

            // Update the bat

            batRectangle.X += (int) (kinect.X * batSpeed);

            if (batRectangle.X < 0) 
                batRectangle.X = 0;
            if (batRectangle.X + batRectangle.Width > GraphicsDevice.Viewport.Width)
                batRectangle.X = GraphicsDevice.Viewport.Width - batRectangle.Width;

            // Check for ball and bat collisions

            if (ballRectangle.Intersects(batRectangle))
            {
                beep.Play();
                ballYspeed = ballYspeed * -1;
            }

            // Check for ball and brick collisions

            for (int i = 0; i < wall.Length; i++)
            {
                if (wall[i].IsVisible)
                {
                    if (wall[i].SpriteRectangle.Intersects(ballRectangle))
                    {
                        // hit a block
                        beep.Play();
                        score = score + wall[i].SpriteScore;
                        wall[i].IsVisible = false;
                        ballYspeed = ballYspeed * -1;

                        blocksLeft--;
                        if (blocksLeft == 0)
                        {
                            wallYPosition += wallYStep;
                            if (wallYPosition > wallYLimit)
                            {
                                wallYPosition = wallYStart;
                                if (ballXspeed > 0)
                                {
                                    ballXspeed += ballSpeedUp;
                                }
                                else
                                {
                                    ballXspeed -= ballSpeedUp;
                                }
                                if (ballYspeed > 0)
                                {
                                    ballYspeed += ballSpeedUp;
                                }
                                else
                                {
                                    ballYspeed -= ballSpeedUp;
                                }
                            }
                            // we have destroyed all the blocks
                            for (int j = 0; j < wall.Length; j++)
                            {
                                wall[j].SpriteRectangle.Y += ballRectangle.Height;
                                wall[j].IsVisible = true;
                                wall[j].SpriteScore += 5;
                            }
                            blocksLeft = wall.Length;
                        }

                        break;
                    }
                }
            }

            for (int i = 0; i < wall2.Length; i++)
            {
                if (wall2[i].IsVisible)
                {
                    if (wall2[i].SpriteRectangle.Intersects(ballRectangle))
                    {
                        // hit a block
                        beep.Play();
                        score = score + wall2[i].SpriteScore;
                        wall2[i].IsVisible = false;
                        ballYspeed = ballYspeed * -1;

                        blocksLeft--;
                        if (blocksLeft == 0)
                        {
                            wallYPosition += wallYStep;
                            if (wallYPosition > wallYLimit)
                            {
                                wallYPosition = wallYStart;
                                if (ballXspeed > 0)
                                {
                                    ballXspeed += ballSpeedUp;
                                }
                                else
                                {
                                    ballXspeed -= ballSpeedUp;
                                }
                                if (ballYspeed > 0)
                                {
                                    ballYspeed += ballSpeedUp;
                                }
                                else
                                {
                                    ballYspeed -= ballSpeedUp;
                                }
                            }
                            // we have destroyed all the blocks
                            for (int j = 0; j < wall.Length; j++)
                            {
                                wall[j].SpriteRectangle.Y += ballRectangle.Height;
                                wall[j].IsVisible = true;
                                wall[j].SpriteScore += 5;
                            }
                            blocksLeft = wall.Length;
                        }

                        break;
                    }
                }
            }

            // Update the score display

            message = "Score: " + score.ToString() + " Lives: " + lives.ToString();

            // Check for losing a life 

            if ((ballRectangle.Y + ballRectangle.Height) >
                 GraphicsDevice.Viewport.Height)
            {
                //loseLife();
            }
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            gamePad1 = GamePad.GetState(PlayerIndex.One);

            // Allows the game to exit
            if (gamePad1.Buttons.Back == ButtonState.Pressed)
                this.Exit();

            keyState = Keyboard.GetState();

            if (keyState.IsKeyDown(Keys.Escape))
                this.Exit();

            kinect.Update();

            switch (state)
            {
                case GameState.error:
                    break;
                case GameState.attractMode:
                    attractUpdate();
                    break;
                case GameState.gamePlaying:
                    gameUpdate();
                    break;
            }

            oldGamePadState = gamePad1;
            oldKeyState = keyState;

            base.Update(gameTime);
        }

        void attractDraw()
        {
            spriteBatch.Draw(ballTexture, ballRectangle, ballColor);
        }

        void drawBlocks()
        {
            for (int i = 0; i < wall.Length; i++)
            {
                if (wall[i].IsVisible)
                {
                    spriteBatch.Draw(
                        wall[i].SpriteTexture,
                        wall[i].SpriteRectangle,
                        wall[i].BlockColour);
                }
            }
            for (int i = 0; i < wall2.Length; i++)
            {
                if (wall2[i].IsVisible)
                {
                    spriteBatch.Draw(
                        wall2[i].SpriteTexture,
                        wall2[i].SpriteRectangle,
                        wall2[i].BlockColour);
                }
            }
        }

        void playDraw()
        {
            drawBlocks();
            spriteBatch.Draw(batTexture, batRectangle, Color.Green);
            spriteBatch.Draw(ballTexture, ballRectangle, ballColor);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.Black);

            spriteBatch.Begin();

            if ( kinect.DepthTexture != null )
                spriteBatch.Draw(kinect.DepthTexture, depthRectangle, depthColor);

            switch (state)
            {
                case GameState.error:
                    break;
                case GameState.attractMode:
                    attractDraw();
                    break;
                case GameState.gamePlaying:
                    playDraw();
                    break;
            }

            // Draw the score message in the middle of the screen
            float messageWidth =
                messageFont.MeasureString(message).X;

            float messageX =
                (GraphicsDevice.Viewport.Width - messageWidth) / 2;

            Vector2 messagePosition = new Vector2(messageX, 10);

            spriteBatch.DrawString(
                messageFont,        // font to use
                message,            // message string
                messagePosition,    // position on display
                Color.Gray);      // colour of text

            messagePosition.X -= 2;
            messagePosition.Y -= 2;

            spriteBatch.DrawString(
                messageFont,        // font to use
                message,            // message string
                messagePosition,    // position on display
                Color.Orange);      // colour of text

            spriteBatch.End();


            base.Draw(gameTime);
        }
    }
}
