﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Threading;
using Microsoft.Kinect;


namespace KinectUtils
{
    /// <summary>
    /// Simple class to provide an interface to the Kinect device
    /// </summary>
    public class KinectManager
    {
        public delegate void ProcessVideoFrame(ColorImageFrame frame);

        public ProcessVideoFrame DoProcessVideoFrame = null;

        public delegate void ProcessDepthFrame(DepthImageFrame frame);

        public ProcessDepthFrame DoProcessDepthFrame = null;

        public delegate void ProcessSkeletons(SkeletonFrame frame);

        public ProcessSkeletons DoProcessSkeletons = null;

        public delegate void StatusChanged(string status);

        public StatusChanged DoStatusChanged = null;

        public KinectSensor Kinect;

        private string statusValue;

        private string status
        {
            get
            {
                return statusValue;
            }
            set
            {
                if (statusValue == value) return;

                statusValue = value;
                if (DoStatusChanged != null)
                {
                    DoStatusChanged(statusValue);
                }
            }
        }

        /// <summary>
        /// Status of the Kinect sensor as a string
        /// </summary>
        public string Status
        {
            get
            {
                return statusValue;
            }
        }

        Thread kinectThread;

        bool kinectActive = true;

        void runKinect()
        {
            status = "Starting..";

            int startTickCount = 0;

            Kinect.Start();

            while (kinectActive)
            {

                if ( Kinect.Status != KinectStatus.Connected ) {
                    status = "Error: " + Kinect.Status.ToString();
                    Thread.Sleep(200);
                    continue;
                }

                if (!Kinect.IsRunning)
                {
                    status = "Starting " + startTickCount.ToString();
                    Thread.Sleep(200);
                    startTickCount++;
                    continue;
                }

                status = "Running";

                if (DoProcessVideoFrame != null)
                {
                    using (ColorImageFrame colorFrame = Kinect.ColorStream.OpenNextFrame(10))
                    {
                        if (colorFrame != null)
                            DoProcessVideoFrame(colorFrame);
                    }
                }

                if (DoProcessDepthFrame != null)
                {
                    using (DepthImageFrame depthFrame = Kinect.DepthStream.OpenNextFrame(10))
                    {
                        if (depthFrame != null)
                            DoProcessDepthFrame(depthFrame);
                    }
                }

                if (DoProcessSkeletons != null)
                {
                    using (SkeletonFrame skeletonFrame = Kinect.SkeletonStream.OpenNextFrame(10))
                    {
                        if (skeletonFrame != null)
                            DoProcessSkeletons(skeletonFrame);
                    }
                }
            }

            status = "Stopping";

            if (Kinect.ColorStream.IsEnabled) Kinect.ColorStream.Disable();
            if (Kinect.DepthStream.IsEnabled) Kinect.DepthStream.Disable();
            if (Kinect.SkeletonStream.IsEnabled) Kinect.SkeletonStream.Disable();

            int stopTickCount = 0;

            Kinect.Stop();

            while (Kinect.IsRunning)
            {
                status = "Stopping " + stopTickCount.ToString();
                Thread.Sleep(200);
                stopTickCount++;
            }

            status = "Stopped";
        }

        /// <summary>
        /// Create a new KinectManager
        /// </summary>
        /// <param name="kinectNo">number of the Kinect in the KinectSensors collection</param>
        /// <param name="colorFormat">required color format. ColorImageFormat.Undefined if no color required</param>
        /// <param name="depthFormat">required depth format. DepthImageFormat.Undefined if no depth required</param>
        /// <param name="skeletonsActive">true if skeleton tracking required</param>
        public KinectManager(int kinectNo, ColorImageFormat colorFormat, DepthImageFormat depthFormat, bool skeletonsActive)
        {
            Kinect = KinectSensor.KinectSensors[kinectNo];

            if (colorFormat != ColorImageFormat.Undefined)
                Kinect.ColorStream.Enable();

            if (depthFormat != DepthImageFormat.Undefined)
                Kinect.DepthStream.Enable(depthFormat);

            if (skeletonsActive)
                Kinect.SkeletonStream.Enable();

        }


        /// <summary>
        /// Create a default sensor with 640x480 depth and video and skeleton active
        /// </summary>
        /// <param name="kinectNo">number of the sensor in the KinectSensors array</param>
        public KinectManager(int kinectNo)
            : this(kinectNo, ColorImageFormat.RgbResolution640x480Fps30, DepthImageFormat.Resolution640x480Fps30, true)
        {
        }

        /// <summary>
        /// Starts the sensor running and generating data
        /// </summary>
        public void Start()
        {
            kinectActive = true;

            kinectThread = new Thread(new ThreadStart(runKinect));

            kinectThread.Start();
        }

        /// <summary>
        /// Stops the sensor
        /// </summary>
        public void Stop()
        {
            kinectActive = false;
        }
    }
}
