﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using System.IO;

using Microsoft.Kinect;
using Microsoft.Speech.AudioFormat;
using Microsoft.Speech.Recognition;

namespace KinectBodyDrawing
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        public MainWindow()
        {
            InitializeComponent();
        }

        #region Speech decoding
        RecognizerInfo kinectRecognizerInfo;
        SpeechRecognitionEngine recognizer;

        KinectSensor myKinect = null;
        KinectAudioSource kinectSource = null;
        Stream audioStream;

        private RecognizerInfo findKinectRecognizerInfo()
        {
            var recognizers = SpeechRecognitionEngine.InstalledRecognizers();

            foreach (RecognizerInfo recInfo in recognizers)
            {
                // look at each recognizer info value to find the one that works for Kinect
                if (recInfo.AdditionalInfo.ContainsKey("Kinect"))
                {
                    string details = recInfo.AdditionalInfo["Kinect"];
                    if (details == "True" && recInfo.Culture.Name == "en-US")
                    {
                        // If we get here we have found the info we want to use
                        return recInfo;
                    }
                }
            }
            return null;
        }

        private void createSpeechEngine()
        {
            kinectRecognizerInfo = findKinectRecognizerInfo();

            if (kinectRecognizerInfo == null)
            {
                MessageBox.Show("Kinect recognizer not found", "Kinect Speech Demo");
                Application.Current.Shutdown();
            }

            try
            {
                recognizer = new SpeechRecognitionEngine(kinectRecognizerInfo);
            }
            catch
            {
                MessageBox.Show("Speech recognition engine could not be loaded", "Kinect Speech Demo");
                Application.Current.Shutdown();
            }
        }

        private void buildCommands()
        {
            Choices commands = new Choices();

            commands.Add("Red");    // set color to red
            commands.Add("Green");  // set color to green
            commands.Add("Blue");   // set color to blue
            commands.Add("Yellow"); // set color to yellow
            commands.Add("Cyan");   // set color to cyan
            commands.Add("Orange"); // set color to orange
            commands.Add("Purple"); // set color to purple
            commands.Add("Magenta");// set color to magenta
            commands.Add("Up");     // lift pen up and stop drawing
            commands.Add("Down");   // put pen down and draw
            commands.Add("Random"); // put draw random colors
            commands.Add("Save");   // save the picture
            commands.Add("Clear");  // clear the screen
            commands.Add("Test");   // draw a test pattern

            GrammarBuilder grammarBuilder = new GrammarBuilder();

            grammarBuilder.Culture = kinectRecognizerInfo.Culture;
            grammarBuilder.Append(commands);

            Grammar grammar = new Grammar(grammarBuilder);

            recognizer.LoadGrammar(grammar);
        }

        private void setupAudio()
        {
            try
            {
                kinectSource = myKinect.AudioSource;
                kinectSource.BeamAngleMode = BeamAngleMode.Adaptive;
                audioStream = kinectSource.Start();
                recognizer.SetInputToAudioStream(audioStream, new SpeechAudioFormatInfo(
                                                      EncodingFormat.Pcm, 16000, 16, 1,
                                                      32000, 2, null));
                recognizer.RecognizeAsync(RecognizeMode.Multiple);
            }
            catch
            {
                MessageBox.Show("Audio stream could not be connected", "Kinect Speech Demo");
                Application.Current.Shutdown();
            }
        }


        private void SetupSpeechRecognition()
        {
            createSpeechEngine();
            buildCommands();
            setupAudio();

            recognizer.SpeechRecognized +=
                new EventHandler<SpeechRecognizedEventArgs>(recognizer_SpeechRecognized);
        }

        private void shutdownSpeechRecognition()
        {
            if (kinectSource != null)
                kinectSource.Stop();

            if (recognizer != null)
            {
                recognizer.RecognizeAsyncCancel();
                recognizer.RecognizeAsyncStop();
            }
        }

        void recognizer_SpeechRecognized(object sender, SpeechRecognizedEventArgs e)
        {
            if (e.Result.Confidence > 0.9f)
            {
                handleCommand(e.Result.Text);
            }
        }

        #endregion

        #region Kinect Skeleton Tracking

        void setupKinect()
        {

            // Check to see if a Kinect is available
            if (KinectSensor.KinectSensors.Count == 0)
            {
                MessageBox.Show("No Kinects detected", "Camera Viewer");
                Application.Current.Shutdown();
            }

            // Get the first Kinect on the computer
            myKinect = KinectSensor.KinectSensors[0];

            // Start the Kinect running and select the depth camera
            try
            {
                myKinect.SkeletonStream.Enable();
                myKinect.Start();
            }
            catch
            {
                MessageBox.Show("Kinect initialise failed", "Camera Viewer");
                Application.Current.Shutdown();
            }

            // connect a handler to the event that fires when new frames are available

            myKinect.SkeletonFrameReady += new EventHandler<SkeletonFrameReadyEventArgs>(myKinect_SkeletonFrameReady);

        }

        bool stopCalled = false;

        void shutdownKinect()
        {
            stopCalled = true;
            if (myKinect != null)
                myKinect.Stop();
        }

        Brush cursorBrush = new SolidColorBrush(Colors.Red);

        Brush skeletonBrush = new SolidColorBrush(Colors.Red);

        void myKinect_SkeletonFrameReady(object sender, SkeletonFrameReadyEventArgs e)
        {
            if (stopCalled) return;

            string message = "No Skeleton Data";
            string qualityMessage = "";

            // Remove the old skeleton
            skeletonCursorCanvas.Children.Clear();
            Brush brush = new SolidColorBrush(Colors.Red);

            Skeleton[] skeletons = null;

            using (SkeletonFrame frame = e.OpenSkeletonFrame())
            {
                if (frame != null)
                {
                    skeletons = new Skeleton[frame.SkeletonArrayLength];
                    frame.CopySkeletonDataTo(skeletons);
                }
            }

            if (skeletons == null) return;

            foreach (Skeleton skeleton in skeletons)
            {
                if (skeleton.TrackingState == SkeletonTrackingState.Tracked)
                {
                    Joint headJoint = skeleton.Joints[JointType.Head];
                    Joint hipCenter = skeleton.Joints[JointType.HipCenter];

                    if (skeleton.ClippedEdges == 0)
                    {
                        qualityMessage = "Good Quality";
                    }
                    else
                    {
                        if ((skeleton.ClippedEdges & FrameEdges.Bottom) != 0)
                            qualityMessage += "Move up\r";
                        if ((skeleton.ClippedEdges & FrameEdges.Top) != 0)
                            qualityMessage += "Move down\r";
                        if ((skeleton.ClippedEdges & FrameEdges.Right) != 0)
                            qualityMessage += "Move left\r";
                        if ((skeleton.ClippedEdges & FrameEdges.Left) != 0)
                            qualityMessage += "Move right ";
                    }

                    draw(skeleton);
                }
            }

            StatusTextBlock.Text = message;
            QualityTextBlock.Text = qualityMessage;
        }

        #endregion

        #region Drawing

        Brush backgroundBrush = new SolidColorBrush(Colors.White);

        Brush drawingBrush = new SolidColorBrush(Colors.Black);

        JointType brushStart = JointType.WristRight;
        JointType brushEnd = JointType.HandRight;

        enum drawMode
        {
            drawingSolidColor,
            drawingRandomColor,
            moving
        }

        drawMode mode = drawMode.moving;

        #region Drawing primitives

        void addLine(Joint j1, Joint j2, Brush brush, Canvas canvas)
        {
            Line boneLine = new Line();
            boneLine.Stroke = brush;
            boneLine.StrokeThickness = 5;

            ColorImagePoint j1P = myKinect.MapSkeletonPointToColor(j1.Position, ColorImageFormat.RgbResolution640x480Fps30);
            boneLine.X1 = j1P.X;
            boneLine.Y1 = j1P.Y;

            ColorImagePoint j2P = myKinect.MapSkeletonPointToColor(j2.Position, ColorImageFormat.RgbResolution640x480Fps30);
            boneLine.X2 = j2P.X;
            boneLine.Y2 = j2P.Y;

            canvas.Children.Add(boneLine);
        }

        #endregion

        void clearCanvas()
        {
            clearCanvas(backgroundBrush);
        }

        void clearCanvas(Brush clearBrush)
        {
            drawCanvas.Children.Clear();

            Rectangle background = new Rectangle();

            background.Width = drawCanvas.Width;
            background.Height = drawCanvas.Height;
            background.Fill = clearBrush;
            drawCanvas.Children.Add(background);
        }

        Random rand = new Random(1);

        Brush randomBrush()
        {
            return new SolidColorBrush(Color.FromRgb(
                (byte)rand.Next(10, 255),
                (byte)rand.Next(10, 255),
                (byte)rand.Next(10, 255)));
        }

        void testDraw()
        {
            clearCanvas(new SolidColorBrush(Colors.Red));

            for (int i = 0; i < 1000; i++)
            {
                Line backBone = new Line();
                backBone.Stroke = randomBrush();
                backBone.StrokeThickness = rand.Next(1, 10);
                backBone.X1 = rand.Next(0, (int)drawCanvas.Width);
                backBone.X2 = rand.Next(0, (int)drawCanvas.Width);

                backBone.Y1 = rand.Next(0, (int)drawCanvas.Height);
                backBone.Y2 = rand.Next(0, (int)drawCanvas.Height);

                drawCanvas.Children.Add(backBone);
            }
        }

        void draw(Skeleton skeleton)
        {
            #region Draw the skeleton cursor

            // Spine
            addLine(skeleton.Joints[JointType.Head], skeleton.Joints[JointType.ShoulderCenter], cursorBrush, skeletonCursorCanvas);
            addLine(skeleton.Joints[JointType.ShoulderCenter], skeleton.Joints[JointType.Spine], cursorBrush, skeletonCursorCanvas);

            // Left leg
            addLine(skeleton.Joints[JointType.Spine], skeleton.Joints[JointType.HipCenter], cursorBrush, skeletonCursorCanvas);
            addLine(skeleton.Joints[JointType.HipCenter], skeleton.Joints[JointType.HipLeft], cursorBrush, skeletonCursorCanvas);
            addLine(skeleton.Joints[JointType.HipLeft], skeleton.Joints[JointType.KneeLeft], cursorBrush, skeletonCursorCanvas);
            addLine(skeleton.Joints[JointType.KneeLeft], skeleton.Joints[JointType.AnkleLeft], cursorBrush, skeletonCursorCanvas);
            addLine(skeleton.Joints[JointType.AnkleLeft], skeleton.Joints[JointType.FootLeft], cursorBrush, skeletonCursorCanvas);

            // Right leg
            addLine(skeleton.Joints[JointType.HipCenter], skeleton.Joints[JointType.HipRight], cursorBrush, skeletonCursorCanvas);
            addLine(skeleton.Joints[JointType.HipRight], skeleton.Joints[JointType.KneeRight], cursorBrush, skeletonCursorCanvas);
            addLine(skeleton.Joints[JointType.KneeRight], skeleton.Joints[JointType.AnkleRight], cursorBrush, skeletonCursorCanvas);
            addLine(skeleton.Joints[JointType.AnkleRight], skeleton.Joints[JointType.FootRight], cursorBrush, skeletonCursorCanvas);

            // Left arm
            addLine(skeleton.Joints[JointType.ShoulderCenter], skeleton.Joints[JointType.ShoulderLeft], cursorBrush, skeletonCursorCanvas);
            addLine(skeleton.Joints[JointType.ShoulderLeft], skeleton.Joints[JointType.ElbowLeft], cursorBrush, skeletonCursorCanvas);
            addLine(skeleton.Joints[JointType.ElbowLeft], skeleton.Joints[JointType.WristLeft], cursorBrush, skeletonCursorCanvas);
            addLine(skeleton.Joints[JointType.WristLeft], skeleton.Joints[JointType.HandLeft], cursorBrush, skeletonCursorCanvas);

            // Right arm
            addLine(skeleton.Joints[JointType.ShoulderCenter], skeleton.Joints[JointType.ShoulderRight], cursorBrush, skeletonCursorCanvas);
            addLine(skeleton.Joints[JointType.ShoulderRight], skeleton.Joints[JointType.ElbowRight], cursorBrush, skeletonCursorCanvas);
            addLine(skeleton.Joints[JointType.ElbowRight], skeleton.Joints[JointType.WristRight], cursorBrush, skeletonCursorCanvas);
            addLine(skeleton.Joints[JointType.WristRight], skeleton.Joints[JointType.HandRight], cursorBrush, skeletonCursorCanvas);

            // Drawing brush
            addLine(skeleton.Joints[brushStart], skeleton.Joints[brushEnd], drawingBrush, skeletonCursorCanvas);

            #endregion

            #region Draw on the canvas

            switch (mode)
            {
                case drawMode.moving:
                    break;

                case drawMode.drawingSolidColor:
                    addLine(skeleton.Joints[brushStart], skeleton.Joints[brushEnd], drawingBrush, drawCanvas);
                    break;

                case drawMode.drawingRandomColor:
                    drawingBrush = randomBrush();
                    addLine(skeleton.Joints[brushStart], skeleton.Joints[brushEnd], drawingBrush, drawCanvas);
                    break;
            }
            #endregion
        }

        void setupDrawing()
        {
            // clear the canvas
            clearCanvas();

            // set the initial drawing color
            setDrawingColor(Colors.Red);

            // Lift the pen up
            penUp();
        }

        #endregion

        #region FileSave

        int frameNo = 0;

        void setupFileName()
        {
            string picturePath = Environment.GetFolderPath(Environment.SpecialFolder.MyPictures);
            string file;

            do
            {
                file = picturePath + @"\Drawing" + frameNo.ToString() + ".jpg";
                if (!File.Exists(file))
                    break;
                frameNo++;
            } while (true);

            fileNameTextBox.Text = picturePath + @"\Drawing" + ".jpg";

        }

        string getFileName()
        {
            if (fileNameTextBox.Text == "")
            {
                Dispatcher.Invoke(new Action(() => { fileSaveStatusLabel.Content = "No filename set"; }));
                return null;
            }

            string destDir = System.IO.Path.GetDirectoryName(fileNameTextBox.Text);
            string destFile = System.IO.Path.GetFileNameWithoutExtension(fileNameTextBox.Text);
            return destDir + @"\" + destFile + frameNo.ToString() + ".jpg";
        }

        void saveBitmapToFile(BitmapSource pictureBitmap, string fileName)
        {
            string message;

            try
            {
                // Save image 
                using (FileStream stream = new FileStream(fileName, FileMode.Create))
                {
                    JpegBitmapEncoder encoder = new JpegBitmapEncoder();
                    encoder.Frames.Add(BitmapFrame.Create(pictureBitmap));
                    encoder.Save(stream);
                }
                message = "Frame " + frameNo + " saved OK";
                frameNo++;
            }
            catch
            {
                message = "Save failed";
            }

            Dispatcher.Invoke(new Action(() => { fileSaveStatusLabel.Content = message; }));
        }

        private void fileNameButton_Click(object sender, RoutedEventArgs e)
        {
            // Configure save file dialog box
            Microsoft.Win32.SaveFileDialog dlg = new Microsoft.Win32.SaveFileDialog();
            dlg.FileName = "Shot"; // Default file name
            dlg.DefaultExt = ".jpg"; // Default file extension
            dlg.Filter = "Pictures (.jpg)|*.jpg"; // Filter files by extension

            // Process save file dialog box results
            if (dlg.ShowDialog() == true)
            {
                // set the filename
                fileNameTextBox.Text = dlg.FileName;
                // set the frame number to 0
                frameNo = 0;
            }
        }

        BitmapSource renderCanvasToBitmap(Canvas canvas)
        {
            // Create a render bitmap and push the canvas to it 

            RenderTargetBitmap renderBitmap =
              new RenderTargetBitmap(
                (int)canvas.Width,
                (int)canvas.Height,
                96,                // dots per inch
                96,                // dots per inch  
                PixelFormats.Pbgra32);

            renderBitmap.Render(canvas);

            return renderBitmap;
        }

        void save()
        {
            string filename = getFileName();

            if (filename != null)
            {
                save(filename);
            }
        }

        void save(string filename)
        {
            BitmapSource bitmap = renderCanvasToBitmap(drawCanvas);
            saveBitmapToFile(bitmap, filename);
        }

        void setupFiles()
        {
            setupFileName();
        }

        #endregion

        #region Commands

        void setDrawingColor(Color color)
        {
            drawingBrush = new SolidColorBrush(color);
            colorTextBlock.Background = drawingBrush;
            penDown();
        }

        void penUp()
        {
            mode = drawMode.moving;
            penTextBlock.Text = "Pen Up";
        }

        void penDown()
        {
            mode = drawMode.drawingSolidColor;
            penTextBlock.Text = "Pen Down";
        }

        void penRandom()
        {
            mode = drawMode.drawingRandomColor;
            penTextBlock.Text = "Random Pen";
            colorTextBlock.Background = new SolidColorBrush(Colors.White);
        }

        void handleCommand(string command)
        {
            commandTextBlock.Text = "Command: " + command;

            switch (command)
            {
                case "Red":
                    setDrawingColor(Colors.Red);
                    break;
                case "Green":
                    setDrawingColor(Colors.Green);
                    break;
                case "Blue":
                    setDrawingColor(Colors.Blue);
                    break;
                case "Yellow":
                    setDrawingColor(Colors.Yellow);
                    break;
                case "Cyan":
                    setDrawingColor(Colors.Cyan);
                    break;
                case "Orange":
                    setDrawingColor(Colors.Orange);
                    break;
                case "Purple":
                    setDrawingColor(Colors.Purple);
                    break;
                case "Magenta":
                    setDrawingColor(Colors.Magenta);
                    break;
                case "Up":
                    penUp();
                    break;
                case "Down":
                    penDown();
                    break;
                case "Random":
                    penRandom();
                    break;
                case "Save":
                    save();
                    break;
                case "Clear":
                    clearCanvas();
                    break;
                case "Test":
                    testDraw();
                    break;
            }
        }

        #endregion

        #region Event Handlers

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            setupKinect();
            SetupSpeechRecognition();
            setupDrawing();
            setupFiles();
        }

        private void saveButton_Click(object sender, RoutedEventArgs e)
        {
            save();
        }

        private void Window_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            shutdownSpeechRecognition();
            shutdownKinect();
        }

        #endregion

    }
}
