﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using System.Threading;
using Microsoft.Kinect;
using System.Media;
using KinectUtils;

namespace MidiMan
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        public MainWindow()
        {
            InitializeComponent();
        }

        #region Kinect Video Display

        byte[] colorData = null;

        WriteableBitmap colorImage = null;
        int frameWidth;
        int frameHeight;

        private static readonly int Bgr32BytesPerPixel = (PixelFormats.Bgr32.BitsPerPixel + 7) / 8;

        void processColorFrame(ColorImageFrame colorFrame)
        {
#if SREENSHOT_PAUSE
            if (trackCount == trackLimit) return;
#endif

            frameWidth = colorFrame.Width;
            frameHeight = colorFrame.Height;
            if (colorData == null)
                colorData = new byte[colorFrame.PixelDataLength];

            colorFrame.CopyPixelDataTo(colorData);
            Dispatcher.Invoke(new Action(() => drawColorFrame()));
        }

        void drawColorFrame()
        {
            if (colorData == null) return;

            if (colorImage == null)
            {
                this.colorImage = new WriteableBitmap(
                    frameWidth,
                    frameHeight,
                    96,  // DpiX
                    96,  // DpiY
                    PixelFormats.Bgr32,
                    null);
                kinectVideoImage.Source = colorImage;
            }

            this.colorImage.WritePixels(
                new Int32Rect(0, 0, frameWidth, frameHeight),
                this.colorData,
                frameWidth * Bgr32BytesPerPixel,
                0);
        }

        #endregion

        #region Kinect Skeleton Display

        Skeleton[] skeletons = null;
        string qualityMessage = "";

        void processSkeletonFrame(SkeletonFrame frame)
        {
            skeletons = new Skeleton[frame.SkeletonArrayLength];
            frame.CopySkeletonDataTo(skeletons);
            Dispatcher.Invoke(new Action(() => drawSkeleton()));
        }

        void drawSkeleton()
        {
            if (skeletons == null) return;

            qualityMessage = "";

            playerCanvas.Children.Clear();

            foreach (Skeleton skeleton in skeletons)
            {
                if (skeleton.TrackingState == SkeletonTrackingState.Tracked)
                {
                    if (skeleton.ClippedEdges == 0)
                    {
                        qualityMessage = "Good Quality";
                    }
                    else
                    {
                        if ((skeleton.ClippedEdges & FrameEdges.Bottom) != 0)
                            qualityMessage += "Move up ";
                        if ((skeleton.ClippedEdges & FrameEdges.Top) != 0)
                            qualityMessage += "Move down ";
                        if ((skeleton.ClippedEdges & FrameEdges.Right) != 0)
                            qualityMessage += "Move left ";
                        if ((skeleton.ClippedEdges & FrameEdges.Left) != 0)
                            qualityMessage += "Move right ";
                    }

                    // Uncomment if you want to draw the bones of the player
                    drawSkeleton(skeleton);

                    playKeyboard(skeleton.Joints[JointType.HandRight]);
                }
            }

            SkeletonQualityTextBlock.Text = qualityMessage;
        }


        Brush skeletonBrush = new SolidColorBrush(Colors.Red);

        void addLine(Joint j1, Joint j2)
        {
            Line boneLine = new Line();
            boneLine.Stroke = skeletonBrush;
            boneLine.StrokeThickness = 5;

            ColorImagePoint j1P = myKinect.MapSkeletonPointToColor(j1.Position, ColorImageFormat.RgbResolution640x480Fps30);
            boneLine.X1 = j1P.X;
            boneLine.Y1 = j1P.Y;

            ColorImagePoint j2P = myKinect.MapSkeletonPointToColor(j2.Position, ColorImageFormat.RgbResolution640x480Fps30);
            boneLine.X2 = j2P.X;
            boneLine.Y2 = j2P.Y;

            playerCanvas.Children.Add(boneLine);
        }

        void drawSkeleton(Skeleton skeleton)
        {
            // Spine
            addLine(skeleton.Joints[JointType.Head], skeleton.Joints[JointType.ShoulderCenter]);
            addLine(skeleton.Joints[JointType.ShoulderCenter], skeleton.Joints[JointType.Spine]);

            // Left leg
            addLine(skeleton.Joints[JointType.Spine], skeleton.Joints[JointType.HipCenter]);
            addLine(skeleton.Joints[JointType.HipCenter], skeleton.Joints[JointType.HipLeft]);
            addLine(skeleton.Joints[JointType.HipLeft], skeleton.Joints[JointType.KneeLeft]);
            addLine(skeleton.Joints[JointType.KneeLeft], skeleton.Joints[JointType.AnkleLeft]);
            addLine(skeleton.Joints[JointType.AnkleLeft], skeleton.Joints[JointType.FootLeft]);

            // Right leg
            addLine(skeleton.Joints[JointType.HipCenter], skeleton.Joints[JointType.HipRight]);
            addLine(skeleton.Joints[JointType.HipRight], skeleton.Joints[JointType.KneeRight]);
            addLine(skeleton.Joints[JointType.KneeRight], skeleton.Joints[JointType.AnkleRight]);
            addLine(skeleton.Joints[JointType.AnkleRight], skeleton.Joints[JointType.FootRight]);

            // Left arm
            addLine(skeleton.Joints[JointType.ShoulderCenter], skeleton.Joints[JointType.ShoulderLeft]);
            addLine(skeleton.Joints[JointType.ShoulderLeft], skeleton.Joints[JointType.ElbowLeft]);
            addLine(skeleton.Joints[JointType.ElbowLeft], skeleton.Joints[JointType.WristLeft]);
            addLine(skeleton.Joints[JointType.WristLeft], skeleton.Joints[JointType.HandLeft]);

            // Right arm
            addLine(skeleton.Joints[JointType.ShoulderCenter], skeleton.Joints[JointType.ShoulderRight]);
            addLine(skeleton.Joints[JointType.ShoulderRight], skeleton.Joints[JointType.ElbowRight]);
            addLine(skeleton.Joints[JointType.ElbowRight], skeleton.Joints[JointType.WristRight]);
            addLine(skeleton.Joints[JointType.WristRight], skeleton.Joints[JointType.HandRight]);

        }

        void playKeyboard(Joint joint)
        {
            if (joint.TrackingState != JointTrackingState.Tracked)
            {
                stopNote();
                return;
            }

            ColorImagePoint notePoint = myKinect.MapSkeletonPointToColor(
                joint.Position, ColorImageFormat.RgbResolution640x480Fps30);

            updateNotes(notePoint.X, notePoint.Y);
        }

        #endregion

        #region Kinect setup

        KinectManager myKinectMgr;
        KinectSensor myKinect;

        void setupKinect()
        {
            try
            {
                myKinectMgr = new KinectManager(
                    0, // Sensor number 0
                    ColorImageFormat.RgbResolution640x480Fps30, // Color image needed
                    DepthImageFormat.Resolution320x240Fps30,    // Depth image needed
                    true);                                      // Skeleton image needed

                myKinect = myKinectMgr.Kinect;

                myKinectMgr.DoProcessVideoFrame +=
                    new KinectManager.ProcessVideoFrame(processColorFrame);
                myKinectMgr.DoProcessSkeletons +=
                    new KinectManager.ProcessSkeletons(processSkeletonFrame);

                myKinectMgr.DoStatusChanged +=
                    new KinectManager.StatusChanged(updateStatus);

                myKinectMgr.Start();
            }
            catch
            {
                MessageBox.Show("Kinect initialise failed", "Bugsplat Game");
                Application.Current.Shutdown();
            }
        }

        void shutdownKinect()
        {
            if (myKinectMgr != null)
                myKinectMgr.Stop();
        }

        #endregion

        #region Status Display

        string newStatus;

        void updateStatus(string status)
        {
            newStatus = status;
            Dispatcher.Invoke(new Action(() => updateMessage()));
        }

        void updateMessage()
        {
            KinectStatusTextBlock.Text = newStatus;
        }

        #endregion

        #region MIDI Control

        MIDIControl myMidi;

        private void setupMidi()
        {
            try
            {
                myMidi = new MIDIControl(0);
            }
            catch
            {
                MessageBox.Show("Midi setup failed", "Midi Control");
                Application.Current.Shutdown();
            }
        }

        int noOfKeys = 13;

        Brush lightKeyBrush = new SolidColorBrush(Colors.Green);
        Brush darkKeyBrush = new SolidColorBrush(Colors.DarkGreen);

        int[] cscaleNoteValues = new int[] { 0, 2, 4, 5, 7, 9, 11 };

        // List of the keys 
        List<TextBlock> keys = new List<TextBlock>();

        private void createKeyboard()
        {
            double keyWidth = keyboardCanvas.Width / noOfKeys;

            for (int i = 0; i < noOfKeys; i++)
            {
                // Make a new key
                TextBlock key = new TextBlock();

                // Set the key size
                key.Width = keyWidth;
                key.Height = keyWidth;

                // Make alternate keys different colors
                if (i % 2 == 0)
                    key.Background = lightKeyBrush;
                else
                    key.Background = darkKeyBrush;

                // Place the key on the keyboard
                keyboardCanvas.Children.Add(key);
                Canvas.SetTop(key, 0);
                Canvas.SetBottom(key, keyWidth);

                Canvas.SetLeft(key, i * keyWidth);
                Canvas.SetRight(key, (i + 1) * keyWidth);

                // tag the key with the key number

                // Find the octave which this note is part of
                int octave = i / 7;
                // find the note number in the octave
                int note = i % 7;

                // an Octave has 12 semitones in it
                key.Tag = 60 + (octave * 12) + cscaleNoteValues[note];

                // put the key in the list
                keys.Add(key);
            }
        }

        bool notePlaying = false;
        int keyNo;

        void updateNotes(int x, int y)
        {
            foreach (TextBlock key in keys)
            {
                // find out if the position is in the key
                if (x < Canvas.GetLeft(key)) continue;
                if (x > Canvas.GetRight(key)) continue;
                if (y < Canvas.GetTop(key)) continue;
                if (y > Canvas.GetBottom(key)) continue;

                // if we get here the position is inside the key

                if (!notePlaying)
                {
                    notePlaying = true;
                    // Extract the key number from the tag
                    keyNo = (int)key.Tag;

                    myMidi.SendMIDIMessage(
                                0x90,       // midi command - note down on channel 0
                                keyNo,      // note value – 60 represents middle C on the keyboard
                                100);       // note velocity – how hard the note was hit in the range 0 to 127

                }
                return;
            }

            // if we get here the position is outside the keys
            // turn off the note

            stopNote();

        }

        void stopNote()
        {
            if (notePlaying)
            {
                notePlaying = false;

                myMidi.SendMIDIMessage(
                                0x80,        // midi command - note up on channel 0
                                keyNo,  // note value – 60 represents middle C on the keyboard
                                100);        // note velocity – how "hard" the note was released in the range 0 to 127
            }
        }

        #endregion

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            setupMidi();
            createKeyboard();
            setupKinect();
        }

        private void Windows_Closing(object sender, EventArgs e)
        {
            if (myKinectMgr != null) myKinectMgr.Stop();
        }

    }
}
