//#define DIAGNOSTICS

using System;
using System.IO.Ports;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;
using GHIElectronics.NETMF.Hardware;
using GHIElectronics.NETMF.FEZ;

using System.Collections;

namespace RobotRemote
{
    /// <summary>
    /// Remote control robot
    /// Receives simple commands from the serial port and acts on them.
    /// Also includes bonus "Cowardly" behaviour.
    /// 
    /// Rob Miles
    /// </summary>
    class RobotDriver
    {
        private FEZ_Pin.AnalogIn DistancePin = FEZ_Pin.AnalogIn.An1;
        private FEZ_Components.DistanceDetector distanceDetector;

        private FEZ_Pin.Digital GreenLedPin = FEZ_Pin.Digital.An4;
        private FEZ_Pin.Digital RedLedPin = FEZ_Pin.Digital.An5;

        private FEZ_Components.LED greenLed;
        private FEZ_Components.LED redLed;

        private FEZ_Pin.Interrupt killSwitchPin = FEZ_Pin.Interrupt.Di4;
        private FEZ_Components.Button killSwitch;

        private FEZ_Pin.Interrupt resetSwitchPin = FEZ_Pin.Interrupt.Di2;
        private FEZ_Components.Button resetSwitch;

        private SerialPort xbeePort;

        byte[] statusBytes;

        private void writeMessage(string message)
        {
            statusBytes = Encoding.UTF8.GetBytes(message);
            xbeePort.Write(statusBytes, 0, statusBytes.Length);
        }

        public void SetupRobot()
        {
            distanceDetector =
                new FEZ_Components.DistanceDetector(
                    DistancePin,
                    FEZ_Components.DistanceDetector.SharpSensorType.GP2D120);

            greenLed = new FEZ_Components.LED(GreenLedPin);
            redLed = new FEZ_Components.LED(RedLedPin);
            greenLed.ShutOff();
            redLed.ShutOff();

            killSwitch = new FEZ_Components.Button(killSwitchPin);
            resetSwitch = new FEZ_Components.Button(resetSwitchPin);

            killSwitch.ButtonPressEvent +=
                new FEZ_Components.Button.ButtonPressEventHandler(killSwitch_ButtonPressEvent);

            resetSwitch.ButtonPressEvent +=
                new FEZ_Components.Button.ButtonPressEventHandler(resetSwitch_ButtonPressEvent);

            xbeePort = FEZ_Components.Fez_Xbee.GetPort(true);

#if DIAGNOSTICS
            setupDiagnostics();
#endif

        }

        #region Kill Switch Behaviours

        void shutDown()
        {
            // This is not thread safe. Need to fix for production code
            FEZMini_Robot.Move(0, 0);
        }

        void restart()
        {
        }

        void resetSwitch_ButtonPressEvent(FEZ_Pin.Interrupt pin, FEZ_Components.Button.ButtonState state)
        {
            if (state == FEZ_Components.Button.ButtonState.Pressed)
            {
                restart();
            }
        }

        void killSwitch_ButtonPressEvent(FEZ_Pin.Interrupt pin, FEZ_Components.Button.ButtonState state)
        {
            if (state == FEZ_Components.Button.ButtonState.Pressed)
            {
                shutDown();
            }
        }

        #endregion

        #region Remote Control Behaviour

        Decoder decoder = UTF8Encoding.UTF8.GetDecoder();

        public void RemoteControl()
        {
            activeCommands += new processCommand(decodeMessage);

            xbeePort.DataReceived += new SerialDataReceivedEventHandler(terminalXbeePort_DataReceived);

            while (true)
            {
                Thread.Sleep(1000);
            }
        }

        private void decodeMessage(string message)
        {
            switch (message)
            {
                case "Forward":
                    redLed.TurnOn();
                    greenLed.TurnOn();
                    FEZMini_Robot.MoveRamp(100, 100, 1);
                    System.Threading.Thread.Sleep(300);
                    FEZMini_Robot.Move(0, 0);
                    redLed.ShutOff();
                    greenLed.ShutOff();
                    writeMessage("<OK>");
                    break;
                case "Back":
                    redLed.TurnOn();
                    greenLed.TurnOn();
                    FEZMini_Robot.MoveRamp(-100, -100, 1);
                    System.Threading.Thread.Sleep(300);
                    FEZMini_Robot.Move(0, 0);
                    redLed.ShutOff();
                    greenLed.ShutOff();
                    writeMessage("<OK>");
                    break;
                case "Stop":
                    redLed.ShutOff();
                    greenLed.ShutOff();
                    FEZMini_Robot.Move(0, 0);
                    System.Threading.Thread.Sleep(300);
                    FEZMini_Robot.Move(0, 0);
                    writeMessage("<OK>");
                    break;
                case "Left":
                    redLed.TurnOn();
                    FEZMini_Robot.MoveRamp(100, -100, 1);
                    System.Threading.Thread.Sleep(300);
                    FEZMini_Robot.Move(0, 0);
                    redLed.ShutOff();
                    writeMessage("<OK>");
                    break;
                case "Right":
                    greenLed.TurnOn();
                    FEZMini_Robot.MoveRamp(-100, 100, 1);
                    System.Threading.Thread.Sleep(300);
                    FEZMini_Robot.Move(0, 0);
                    greenLed.ShutOff();
                    writeMessage("<OK>");
                    break;
                case "Distance":
                    float distance = distanceDetector.GetDistance_cm();
                    writeMessage("<" + distance.ToString() + ">");
                    writeMessage("<OK>");
                    break;
                default:
                    writeMessage("<Command: " + message + " not recognised>");
                    break;
            }
        }

        #endregion

        #region Command Transport

        /// <summary>
        /// Bound to the arrival of serial port data. Pumps incoming characters into ProcessCh
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void terminalXbeePort_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            byte[] singleByteBuffer = new byte[1];

            if (e.EventType == SerialData.Chars)
            {
                while (xbeePort.BytesToRead > 0) {
                    xbeePort.Read(singleByteBuffer, 0, 1);
                    processCh((char)singleByteBuffer[0]);
                }
            }
        }

        /// <summary>
        ///  Characters to flag the start and end of a message
        /// </summary>
        char startCh = '<';
        char endCh = '>';

        /// <summary>
        /// We are either waiting for a start character or building the command
        /// </summary>
        enum ReadStates
        {
            waiting,
            buildingCommand
        }

        ReadStates state;

        /// <summary>
        /// Maximum length of a command verb. Anything longer than this is ignored.
        /// </summary>
        const int bufferLength = 50;

        /// <summary>
        /// Buffer to hold the command string
        /// </summary>
        char[] commandBuffer = new char[bufferLength];

        /// <summary>
        /// Write position for incoming command characters
        /// </summary>
        int bufferPos = 0;

        /// <summary>
        ///  List of commands received and being consumed
        /// </summary>
        ArrayList commands = new ArrayList();

        /// <summary>
        /// Processes incoming characters and assembles a command
        /// Command has the form <message> where message is the command message
        /// </summary>
        /// <param name="ch">character to be addedto the command</param>
        void processCh ( char ch )
        {
            switch (state)
            {
                case ReadStates.waiting:
                    if ( ch == startCh )
                    {
                        bufferPos = 0;
                        state = ReadStates.buildingCommand;
                    }
                    break;

                case ReadStates.buildingCommand:
                    if (ch == endCh)
                    {
                        string command = new String(commandBuffer, 0, bufferPos);

                        Monitor.Enter(commandLock);
                        commands.Add(command);
                        Monitor.Exit(commandLock);

                        Thread x = new Thread(new ThreadStart(processCommands));
                        x.Start();

                        state = ReadStates.waiting;
                    }
                    else
                    {
                        if (bufferPos < bufferLength)
                        {
                            commandBuffer[bufferPos] = ch;
                            bufferPos++;
                        }
                    }
                    break;
            }
        }

        private object commandLock = new object();

        delegate void processCommand(string command);

        processCommand activeCommands = null;

        void processCommands() 
        {

            Monitor.Enter(commandLock);
            string command = commands[0] as string;
            commands.RemoveAt(0);
            Monitor.Exit(commandLock);

            if (activeCommands != null)
            {
                activeCommands(command);
            }
        }

        #endregion


        #region Performance Test Code

        int messageCount = 0;

        void performanceTest(string command)
        {
            switch (command)
            {
                case "reset": 
                    messageCount = 0;
                    break;

                default:
                    int messageNo=0;
                    try
                    {
                        messageNo = int.Parse(command);
                    }
                    catch
                    {
                        Debug.Print ("Invalid number: " + command);
                        return;
                    }

                    if (messageNo != messageCount)
                    {
                        Debug.Print("Failed: " + messageNo + " " + messageCount);
                        return;
                    }
                    messageCount++;
                    break;
            }
        }

        #endregion

    }
}
